@@ -16,6 +16,16 @@ import os
 
 from typing import Any, Callable, Dict, List, Optional, Set, cast
 
+from .changes_resolvers import (
+    FragmentResolver,
+    PluginResolver,
+    LegacyFragmentResolver,
+    LegacyPluginResolver,
+    LegacyObjectResolver,
+    ChangesDataFragmentResolver,
+    ChangesDataPluginResolver,
+    ChangesDataObjectResolver
+)
 from .config import ChangelogConfig
 from .fragment import ChangelogFragment, load_fragments
 from .logger import LOGGER
@@ -25,39 +35,6 @@ from .utils import get_version_constructor, is_release_version
 from .yaml import load_yaml, store_yaml
 
 
-class FragmentResolver(metaclass=abc.ABCMeta):
-    # pylint: disable=too-few-public-methods
-    """
-    Allows to resolve a release section to a list of changelog fragments.
-    """
-
-    @abc.abstractmethod
-    def resolve(self, release: dict) -> List[ChangelogFragment]:
-        """
-        Return a list of ``ChangelogFragment`` objects from the given release object.
-
-        :arg release: A release description
-        :return: A list of changelog fragments
-        """
-
-
-class PluginResolver(metaclass=abc.ABCMeta):
-    # pylint: disable=too-few-public-methods
-    """
-    Allows to resolve a release section to a plugin description database.
-    """
-
-    @abc.abstractmethod
-    def resolve(self, release: dict) -> Dict[str, List[Dict[str, Any]]]:
-        """
-        Return a dictionary of plugin types mapping to lists of plugin descriptions
-        for the given release.
-
-        :arg release: A release description
-        :return: A map of plugin types to lists of plugin descriptions
-        """
-
-
 class ChangesBase(metaclass=abc.ABCMeta):
     """
     Read, write and manage change metadata.
@@ -67,6 +44,7 @@ class ChangesBase(metaclass=abc.ABCMeta):
     path: str
     data: dict
     known_plugins: Set[str]
+    known_objects: Set[str]
     known_fragments: Set[str]
     ancestor: Optional[str]
 
@@ -76,6 +54,7 @@ class ChangesBase(metaclass=abc.ABCMeta):
         self.data = self.empty()
         self.known_fragments = set()
         self.known_plugins = set()
+        self.known_objects = set()
         self.ancestor = None
 
     def version_constructor(self, version: str) -> Any:
@@ -140,6 +119,14 @@ class ChangesBase(metaclass=abc.ABCMeta):
         of plugins.
         """
 
+    @abc.abstractmethod
+    def update_objects(self, objects: List[PluginDescription],
+                       allow_removals: Optional[bool]) -> None:
+        """
+        Update object descriptions, and remove objects which are not in the provided list
+        of objects.
+        """
+
     @abc.abstractmethod
     def update_fragments(self, fragments: List[ChangelogFragment],
                          load_extra_fragments: Optional[
@@ -164,18 +151,20 @@ class ChangesBase(metaclass=abc.ABCMeta):
         self.data['ancestor'] = self.ancestor
         store_yaml(self.path, self.data)
 
-    def add_release(self, version: str, codename: Optional[str], release_date: datetime.date):
+    def add_release(self, version: str, codename: Optional[str],
+                    release_date: datetime.date, update_existing=False):
         """
         Add a new releases to the changes metadata.
         """
         if version not in self.releases:
-            self.releases[version] = dict(
-                release_date=release_date.isoformat(),
-            )
-            if codename is not None:
-                self.releases[version]['codename'] = codename
-        else:
+            self.releases[version] = dict()
+        elif not update_existing:
             LOGGER.warning('release {} already exists', version)
+            return
+
+        self.releases[version]['release_date'] = release_date.isoformat()
+        if codename is not None:
+            self.releases[version]['codename'] = codename
 
     @abc.abstractmethod
     def add_fragment(self, fragment: ChangelogFragment, version: str):
@@ -222,6 +211,34 @@ class ChangesBase(metaclass=abc.ABCMeta):
 
         return True
 
+    def add_object(self, ansible_object: PluginDescription, version: str):
+        """
+        Add a new object to the change metadata for the given version.
+
+        If the object happens to be already known (for another version),
+        it will not be added.
+
+        :return: ``True`` if the object was added for this version
+        """
+        composite_name = '%s/%s' % (ansible_object.type, ansible_object.name)
+
+        if composite_name in self.known_objects:
+            return False
+
+        self.known_objects.add(composite_name)
+
+        if 'objects' not in self.releases[version]:
+            self.releases[version]['objects'] = {}
+
+        objects = self.releases[version]['objects']
+
+        if ansible_object.type not in objects:
+            objects[ansible_object.type] = []
+
+        objects[ansible_object.type].append(self._create_plugin_entry(ansible_object))
+
+        return True
+
     @abc.abstractmethod
     def get_plugin_resolver(
             self, plugins: Optional[List[PluginDescription]] = None) -> PluginResolver:
@@ -231,6 +248,12 @@ class ChangesBase(metaclass=abc.ABCMeta):
         If the plugins are not provided and needed by this object, they might be loaded.
         """
 
+    @abc.abstractmethod
+    def get_object_resolver(self) -> PluginResolver:
+        """
+        Create a object resolver.
+        """
+
     @abc.abstractmethod
     def get_fragment_resolver(
             self, fragments: Optional[List[ChangelogFragment]] = None) -> FragmentResolver:
@@ -241,82 +264,6 @@ class ChangesBase(metaclass=abc.ABCMeta):
         """
 
 
-class LegacyFragmentResolver(FragmentResolver):
-    # pylint: disable=too-few-public-methods
-    """
-    Given a list of changelog fragments, allows to resolve from a list of fragment names.
-    """
-
-    fragments: Dict[str, ChangelogFragment]
-
-    def __init__(self, fragments: List[ChangelogFragment]):
-        """
-        Create a simple fragment resolver.
-        """
-        self.fragments = dict()
-        for fragment in fragments:
-            self.fragments[fragment.name] = fragment
-
-    def resolve(self, release: dict) -> List[ChangelogFragment]:
-        """
-        Return a list of ``ChangelogFragment`` objects from the given release object.
-
-        :arg release: A release description
-        :return: A list of changelog fragments
-        """
-        fragment_names: List[str] = release.get('fragments', [])
-        return [self.fragments[fragment] for fragment in fragment_names]
-
-
-class LegacyPluginResolver(PluginResolver):
-    # pylint: disable=too-few-public-methods
-    """
-    Provides a plugin resolved based on a list of ``PluginDescription`` objects.
-    """
-
-    plugins: Dict[str, Dict[str, Dict[str, Any]]]
-
-    @staticmethod
-    def resolve_plugin(plugin: PluginDescription) -> Dict[str, Any]:
-        """
-        Convert a ``PluginDecscription`` object to a plugin description dictionary.
-        """
-        return dict(
-            name=plugin.name,
-            namespace=plugin.namespace,
-            description=plugin.description,
-        )
-
-    def __init__(self, plugins: List[PluginDescription]):
-        """
-        Create a simple plugin resolver from a list of ``PluginDescription`` objects.
-        """
-        self.plugins = dict()
-        for plugin in plugins:
-            if plugin.type not in self.plugins:
-                self.plugins[plugin.type] = dict()
-
-            self.plugins[plugin.type][plugin.name] = self.resolve_plugin(plugin)
-
-    def resolve(self, release: dict) -> Dict[str, List[Dict[str, Any]]]:
-        """
-        Return a dictionary of plugin types mapping to lists of plugin descriptions
-        for the given release.
-
-        :arg release: A release description
-        :return: A map of plugin types to lists of plugin descriptions
-        """
-        result = dict()
-        if 'modules' in release:
-            result['module'] = [self.plugins['module'][module_name]
-                                for module_name in release['modules']]
-        if 'plugins' in release:
-            for plugin_type, plugin_names in release['plugins'].items():
-                result[plugin_type] = [self.plugins[plugin_type][plugin_name]
-                                       for plugin_name in plugin_names]
-        return result
-
-
 class ChangesMetadata(ChangesBase):
     """
     Read, write and manage classic Ansible (2.9 and earlier) change metadata.
@@ -326,14 +273,14 @@ class ChangesMetadata(ChangesBase):
         """
         Create legacy change metadata.
         """
-        super(ChangesMetadata, self).__init__(config, path)
+        super().__init__(config, path)
         self.load()
 
     def load(self, data_override: Optional[dict] = None) -> None:
         """
         Load the change metadata from disk.
         """
-        super(ChangesMetadata, self).load(data_override=data_override)
+        super().load(data_override=data_override)
 
         for _, config in self.releases.items():
             for plugin_type, plugin_names in config.get('plugins', {}).items():
@@ -384,6 +331,14 @@ class ChangesMetadata(ChangesBase):
                     self.known_plugins -= set(
                         '%s/%s' % (plugin_type, plugin) for plugin in invalid_plugins)
 
+    def update_objects(self, objects: List[PluginDescription],
+                       allow_removals: Optional[bool]) -> None:
+        """
+        Update object descriptions, and remove objects which are not in the provided list
+        of objects.
+        """
+        return
+
     def update_fragments(self, fragments: List[ChangelogFragment],
                          load_extra_fragments: Optional[
                              Callable[[str], List[ChangelogFragment]]] = None
@@ -451,6 +406,12 @@ class ChangesMetadata(ChangesBase):
                                    force_reload=False)
         return LegacyPluginResolver(plugins)
 
+    def get_object_resolver(self) -> PluginResolver:
+        """
+        Create a dummy object resolver.
+        """
+        return LegacyObjectResolver()
+
     def get_fragment_resolver(
             self, fragments: Optional[List[ChangelogFragment]] = None) -> FragmentResolver:
         """
@@ -463,47 +424,6 @@ class ChangesMetadata(ChangesBase):
         return LegacyFragmentResolver(fragments)
 
 
-class ChangesDataFragmentResolver(FragmentResolver):
-    # pylint: disable=too-few-public-methods
-    """
-    A ``FragmentResolver`` class for modern ``ChangesData`` objects.
-    """
-
-    def resolve(self, release: dict) -> List[ChangelogFragment]:
-        """
-        Return a list of ``ChangelogFragment`` objects from the given release object.
-
-        :arg release: A release description
-        :return: A list of changelog fragments
-        """
-        changes = release.get('changes')
-        if changes is None:
-            return []
-        return [ChangelogFragment.from_dict(changes)]
-
-
-class ChangesDataPluginResolver(PluginResolver):
-    # pylint: disable=too-few-public-methods
-    """
-    A ``PluginResolver`` class for modern ``ChangesData`` objects.
-    """
-
-    def resolve(self, release: dict) -> Dict[str, List[Dict[str, Any]]]:
-        """
-        Return a dictionary of plugin types mapping to lists of plugin descriptions
-        for the given release.
-
-        :arg release: A release description
-        :return: A map of plugin types to lists of plugin descriptions
-        """
-        result = dict()
-        if 'modules' in release:
-            result['module'] = release['modules']
-        if 'plugins' in release:
-            result.update(release['plugins'])
-        return result
-
-
 class ChangesData(ChangesBase):
     """
     Read, write and manage modern change metadata.
@@ -519,7 +439,7 @@ class ChangesData(ChangesBase):
 
         :arg data_override: Allows to load data from dictionary instead from disk
         """
-        super(ChangesData, self).__init__(config, path)
+        super().__init__(config, path)
         self.config = config
         self.load(data_override=data_override)
 
@@ -527,13 +447,17 @@ class ChangesData(ChangesBase):
         """
         Load the change metadata from disk.
         """
-        super(ChangesData, self).load(data_override=data_override)
+        super().load(data_override=data_override)
 
         for _, config in self.releases.items():
             for plugin_type, plugins in config.get('plugins', {}).items():
                 self.known_plugins |= set(
                     '%s/%s' % (plugin_type, plugin['name']) for plugin in plugins)
 
+            for object_type, objects in config.get('objects', {}).items():
+                self.known_objects |= set(
+                    '%s/%s' % (object_type, ansible_object['name']) for ansible_object in objects)
+
             modules = config.get('modules', [])
 
             self.known_plugins |= set('module/%s' % module['name'] for module in modules)
@@ -590,6 +514,40 @@ class ChangesData(ChangesBase):
                             plugin
                             for plugin in config['plugins'][plugin_type]]
 
+    def update_objects(self, objects: List[PluginDescription],
+                       allow_removals: Optional[bool]) -> None:
+        """
+        Update object descriptions, and remove objects which are not in the provided list
+        of objects.
+        """
+        valid_objects: Dict[str, Dict[str, PluginDescription]] = collections.defaultdict(dict)
+
+        for ansible_object in objects:
+            valid_objects[ansible_object.type][ansible_object.name] = ansible_object
+
+        for _, config in self.releases.items():
+            if 'objects' in config:
+                for object_type in config['objects']:
+                    invalid_object_names = set(
+                        ansible_object['name'] for ansible_object in config['objects'][object_type]
+                        if ansible_object['name'] not in valid_objects[object_type])
+                    if allow_removals:
+                        config['objects'][object_type] = [
+                            self._create_plugin_entry(
+                                valid_objects[object_type][ansible_object['name']])
+                            for ansible_object in config['objects'][object_type]
+                            if ansible_object['name'] not in invalid_object_names]
+                        self.known_objects -= set(
+                            '%s/%s' % (object_type, object_name)
+                            for object_name in invalid_object_names)
+                    else:
+                        config['objects'][object_type] = [
+                            self._create_plugin_entry(
+                                valid_objects[object_type][ansible_object['name']])
+                            if ansible_object['name'] not in invalid_object_names else
+                            ansible_object
+                            for ansible_object in config['objects'][object_type]]
+
     def update_fragments(self, fragments: List[ChangelogFragment],
                          load_extra_fragments: Optional[
                              Callable[[str], List[ChangelogFragment]]] = None
@@ -621,13 +579,14 @@ class ChangesData(ChangesBase):
                 self.known_fragments -= invalid_fragments
 
                 config['changes'] = ChangelogFragment.combine([
-                    valid_fragments[fragment] for fragment in config['fragments']])
+                    valid_fragments[fragment] for fragment in config['fragments']],
+                    ignore_obj_adds=True)
 
     def sort(self) -> None:
         """
         Sort change metadata in place.
         """
-        super(ChangesData, self).sort()
+        super().sort()
 
         for _, config in self.data['releases'].items():
             if 'modules' in config:
@@ -638,6 +597,12 @@ class ChangesData(ChangesBase):
                     config['plugins'][plugin_type] = sorted(
                         config['plugins'][plugin_type], key=lambda plugin: plugin['name'])
 
+            if 'objects' in config:
+                for object_type in config['objects']:
+                    config['objects'][object_type] = sorted(
+                        config['objects'][object_type],
+                        key=lambda ansible_object: ansible_object['name'])
+
             if 'fragments' in config:
                 config['fragments'] = sorted(config['fragments'])
 
@@ -647,11 +612,85 @@ class ChangesData(ChangesBase):
                     for section, entries in sorted(config['changes'].items())
                 }
 
+    def _add_fragment_obj(self, obj_class, obj_type, entry, version: str):
+        """
+        Add an object or a plugin from a changelog fragment.
+        """
+        if obj_class == 'object':
+            composite_name = '%s/%s' % (obj_type, entry['name'])
+            if composite_name in self.known_objects:
+                LOGGER.warning(
+                    'Ignore adding %s object "%s" from changelog fragment' % (
+                        obj_type, entry['name']))
+                return
+            self.known_objects.add(composite_name)
+            toplevel_type = 'objects'
+            has_categories = True
+        if obj_class == 'plugin':
+            composite_name = '%s/%s' % (obj_type, entry['name'])
+            if composite_name in self.known_plugins:
+                LOGGER.warning(
+                    'Ignore adding %s plugin "%s" from changelog fragment' % (
+                        obj_type, entry['name']))
+                return
+            self.known_plugins.add(composite_name)
+            if obj_type == 'module':
+                toplevel_type = 'modules'
+                has_categories = False
+            else:
+                toplevel_type = 'plugins'
+                has_categories = True
+
+        if has_categories:
+            if toplevel_type not in self.releases[version]:
+                self.releases[version][toplevel_type] = {}
+
+            categories = self.releases[version][toplevel_type]
+
+            if obj_type not in categories:
+                categories[obj_type] = []
+
+            obj_list = categories[obj_type]
+        else:
+            if toplevel_type not in self.releases[version]:
+                self.releases[version][toplevel_type] = []
+
+            obj_list = self.releases[version][toplevel_type]
+
+        obj_list.append({
+            'name': entry['name'],
+            'description': entry['description'],
+            'namespace': entry.get('namespace'),
+        })
+
+    def _add_fragment_content(self, version: str, changes: Dict[str, Any],
+                              section: str, lines: Any):
+        """
+        Add contents of a changelog fragment. Helps implementing add_fragment().
+        """
+        if section == self.config.prelude_name:
+            if section in changes:
+                raise ValueError('Found prelude section "{0}" more than once!'.format(section))
+            changes[section] = lines
+        elif section == self.config.trivial_section_name:
+            # Ignore trivial section entries
+            return
+        elif section.startswith('add') and '.' in section:
+            obj_class, obj_type = section[4:].split('.', 1)
+            for entry in lines:
+                self._add_fragment_obj(obj_class, obj_type, entry, version)
+        elif section not in self.config.sections:
+            raise ValueError('Found unknown section "{0}"'.format(section))
+        else:
+            if section not in changes:
+                changes[section] = []
+            changes[section].extend(lines)
+
     def add_fragment(self, fragment: ChangelogFragment, version: str):
         """
         Add a changelog fragment to the change metadata.
         """
-        if fragment.name in self.known_fragments:
+        if fragment.name in self.known_fragments and self.config.prevent_known_fragments:
             return False
 
         self.known_fragments.add(fragment.name)
@@ -664,16 +703,7 @@ class ChangesData(ChangesBase):
             self.releases[version]['fragments'] = []
 
         for section, lines in fragment.content.items():
-            if section == self.config.prelude_name:
-                if section in changes:
-                    raise ValueError('Found prelude section "{0}" more than once!'.format(section))
-                changes[section] = lines
-            elif section not in self.config.sections:
-                raise ValueError('Found unknown section "{0}"'.format(section))
-            else:
-                if section not in changes:
-                    changes[section] = []
-                changes[section].extend(lines)
+            self._add_fragment_content(version, changes, section, lines)
 
         self.releases[version]['fragments'].append(fragment.name)
         return True
@@ -691,6 +721,12 @@ class ChangesData(ChangesBase):
         """
         return ChangesDataPluginResolver()
 
+    def get_object_resolver(self) -> PluginResolver:
+        """
+        Create a object resolver.
+        """
+        return ChangesDataObjectResolver()
+
     def get_fragment_resolver(
             self, fragments: Optional[List[ChangelogFragment]] = None) -> FragmentResolver:
         """
@@ -767,24 +803,114 @@ def load_changes(config: ChangelogConfig) -> ChangesBase:
     return ChangesData(config, path)
 
 
-def add_release(config: ChangelogConfig,  # pylint: disable=too-many-arguments
+def _filter_version_exact(version: str, version_added: Optional[str]) -> bool:
+    return bool(version_added) and any([
+        version.startswith('%s.' % version_added),
+        version.startswith('%s-' % version_added),  # needed for semver
+        version.startswith('%s+' % version_added),  # needed for semver
+        version == version_added
+    ])
+
+
+def _create_filter_version_range(prev_version: str, current_version: str, config: ChangelogConfig
+                                 ) -> Callable[[Optional[str]], bool]:
+    version_constructor = get_version_constructor(config)
+    prev_version_ = version_constructor(prev_version)
+    current_version_ = version_constructor(current_version)
+
+    def f(version_added: Optional[str]) -> bool:
+        if not version_added:
+            return False
+        version_added_ = version_constructor(version_added)
+        return prev_version_ < version_added_ <= current_version_
+
+    return f
+
+
+def _add_plugins_filters(changes: ChangesBase,
+                         plugins: List[PluginDescription],
+                         objects: List[PluginDescription],
+                         version: str,
+                         prev_version: Optional[str],
+                         ) -> None:
+    if prev_version is None:
+        def version_filter(obj: PluginDescription) -> bool:
+            return _filter_version_exact(version, obj.version_added)
+    else:
+        inner_version_filter = _create_filter_version_range(prev_version, version, changes.config)
+
+        def version_filter(obj: PluginDescription) -> bool:
+            return inner_version_filter(obj.version_added)
+
+    # filter out plugins which were not added in this release
+    plugins = list(filter(version_filter, plugins))
+
+    # filter out objects which were not added in this release
+    objects = list(filter(version_filter, objects))
+
+    for plugin in plugins:
+        changes.add_plugin(plugin, version)
+
+    for ansible_object in objects:
+        changes.add_object(ansible_object, version)
+
+
+def _add_fragments(changes: ChangesBase,
+                   fragments: List[ChangelogFragment],
+                   version: str,
+                   show_release_summary_warning: bool
+                   ) -> List[ChangelogFragment]:
+    fragments_added = []
+    has_release_summary = False
+    for fragment in fragments:
+        if changes.add_fragment(fragment, version):
+            fragments_added.append(fragment)
+            if changes.config.prelude_name in fragment.content:
+                has_release_summary = True
+
+    if not has_release_summary and show_release_summary_warning:
+        LOGGER.warning(
+            'Found no {} section in the changelog for this release. While this is not required,'
+            ' we suggest to add one with basic information on the release.'.format(
+                changes.config.prelude_name))
+
+    return fragments_added
+
+
+def add_release(config: ChangelogConfig,  # pylint: disable=too-many-arguments,too-many-locals
                 changes: ChangesBase,
                 plugins: List[PluginDescription],
                 fragments: List[ChangelogFragment],
                 version: str,
                 codename: Optional[str],
                 date: datetime.date,
-                save_changes: bool = True) -> None:
+                save_changes: bool = True,
+                update_existing: bool = False,
+                objects: Optional[List[PluginDescription]] = None,
+                prev_version: Optional[str] = None,
+                show_release_summary_warning: bool = True,
+                ) -> None:
     """
     Add a release to the change metadata.
 
     :arg changes: Changes metadata to update
     :arg plugins: List of all plugin descriptions
+    :arg objects: List of all object descriptions
     :arg fragments: List of all changelog fragments
     :arg version: The version for the new release
     :arg codename: The codename for the new release. Optional for collections
     :arg date: The release date
+    :arg update_existing: When set to ``True``, will update an existing release
+                          instead of ignoring it
+    :arg prev_version: When provided, all plugins added after prev_version are included, and not
+                       only the ones added in this version.
+    :arg show_release_summary_warning: When set to ``True``, show a warning when the release has
+                                       no ``release_summary``.
     """
+    # for backwards compatibility, objects can be None
+    if objects is None:
+        objects = []
+
     # make sure the version parses
     version_constructor = get_version_constructor(config)
     version_constructor(version)
@@ -792,23 +918,11 @@ def add_release(config: ChangelogConfig,  # pylint: disable=too-many-arguments
     LOGGER.info('release version {} is a {} version', version,
                 'release' if is_release_version(config, version) else 'pre-release')
 
-    # filter out plugins which were not added in this release
-    plugins = list(filter(lambda p: any([
-        version.startswith('%s.' % p.version_added),
-        version.startswith('%s-' % p.version_added),  # needed for semver
-        version.startswith('%s+' % p.version_added),  # needed for semver
-        version == p.version_added
-    ]), plugins))
+    changes.add_release(version, codename, date, update_existing=update_existing)
 
-    changes.add_release(version, codename, date)
+    _add_plugins_filters(changes, plugins, objects, version, prev_version)
 
-    for plugin in plugins:
-        changes.add_plugin(plugin, version)
-
-    fragments_added = []
-    for fragment in fragments:
-        if changes.add_fragment(fragment, version):
-            fragments_added.append(fragment)
+    fragments_added = _add_fragments(changes, fragments, version, show_release_summary_warning)
 
     if save_changes:
         changes.save()
